




typedef struct {
	char text[3];
} Cpu_Memory_Cell;

#define NUM_MEMORY_CELLS 256
Cpu_Memory_Cell memory[NUM_MEMORY_CELLS];
_Static_assert(NUM_MEMORY_CELLS % 4 == 0, "NUM_MEMORY_CELLS must be a multiple of 4.");



#if 0
typedef struct {
	char text[3];
} Cpu_Register;
#endif

typedef Cpu_Memory_Cell Cpu_Register;

#define NUM_REGISTERS 8
Cpu_Register program_counter;// = {"00"};
char *register_names[NUM_REGISTERS] = {
	"Register 0", "Register 1", "Register 2", "Register 3",
	"Register 4", "Register 5", "Register 6", "Register 7",
};
Cpu_Register registers[NUM_REGISTERS];

bool cpu_running = false;

void cpu_memory_clear() {
	for(int i = 0; i < NUM_MEMORY_CELLS; i++)
	{
		memory[i].text[0] = '0';
		memory[i].text[1] = '0';
	}
	for(int i = 0; i < NUM_REGISTERS; i++) {
		registers[i].text[0] = '0';
		registers[i].text[1] = '0';
	}
	program_counter.text[0] = '0';
	program_counter.text[1] = '0';
}


void cpu_reset() {
	for(int i = 0; i < NUM_REGISTERS; i++) {
		registers[i].text[0] = '0';
		registers[i].text[1] = '0';
	}
	program_counter.text[0] = '0';
	program_counter.text[1] = '0';
	cpu_running = true;
}

int char_to_nibble(char c) {
	char nibble = c;
	if(nibble >= 'A') {
		nibble -= 7;
	}
	nibble -= '0';
	return nibble;
}

int cell_to_value(Cpu_Memory_Cell cell) {
	char ms_nibble = char_to_nibble(cell.text[0]);
	char ls_nibble = char_to_nibble(cell.text[1]);
	int value = ms_nibble << 4 | ls_nibble;
	return value;
}

void cell_copy(Cpu_Memory_Cell *dst, Cpu_Memory_Cell src) {
	dst->text[0] = src.text[0];
	dst->text[1] = src.text[1];
}

void cell_assign_value(Cpu_Memory_Cell *dst, int value) {
	assert(value <= 255);
	snprintf(dst->text, sizeof(dst->text), "%2.2X", value);
}

typedef enum {
	CPU_OPCODE_NONE  = 0x00, // X X X
	CPU_OPCODE_HALT  = 0x01, // X X X
	CPU_OPCODE_LOAD  = 0x02, // X R R
	CPU_OPCODE_STORE = 0x03, // X R R
	CPU_OPCODE_SET   = 0x04, // X R I
	CPU_OPCODE_MOV   = 0x05, // X R R
	CPU_OPCODE_JUMP  = 0x06, // X X R
	CPU_OPCODE_ADD   = 0x07, // R R R
	CPU_OPCODE_SUB   = 0x08, // R R R
} Cpu_Opcode;
void cpu_step() {
	if(!cpu_running) {
		return;
	}

	int pc = cell_to_value(program_counter);
	if(pc > NUM_MEMORY_CELLS - 3) {
		cpu_running = false;
		return;
	}
	int opcode = cell_to_value(memory[pc]);
	int param1 = cell_to_value(memory[pc+1]);
	int param2 = cell_to_value(memory[pc+2]);
	int param3 = cell_to_value(memory[pc+3]);
	printf("Step params: %2.2X  %2.2X  %2.2X\n", param1, param2, param3);
	int new_pc = pc + 4;
	switch(opcode) {
		case CPU_OPCODE_NONE: break;
		case CPU_OPCODE_HALT: {
			cpu_running = 0;
			new_pc = pc;
		} break;
		case CPU_OPCODE_LOAD: {
			cell_copy(&registers[param2], memory[param3]);
		} break;
		case CPU_OPCODE_STORE: {
			cell_copy(&memory[param3], registers[param2]);
		} break;
		case CPU_OPCODE_SET: {
			cell_copy(&registers[param2], memory[pc+3]);
		} break;
		case CPU_OPCODE_MOV: {
			cell_copy(&registers[param2], registers[param3]);
		} break;
		case CPU_OPCODE_JUMP: {
			new_pc = cell_to_value(registers[param3]);
		} break;
		case CPU_OPCODE_ADD: {
			int left_operand  = cell_to_value(registers[param2]);
			int right_operand = cell_to_value(registers[param3]);
			int result = left_operand + right_operand;
			cell_assign_value(&registers[param1], result);
		} break;
		case CPU_OPCODE_SUB: {
			int left_operand  = cell_to_value(registers[param2]);
			int right_operand = cell_to_value(registers[param3]);
			int result = left_operand - right_operand;
			cell_assign_value(&registers[param1], result);
		} break;
	}

	if(new_pc > NUM_MEMORY_CELLS - 3) {
		cpu_running = false;
		return;
	} else {
		cell_assign_value(&program_counter, new_pc);
	}
}








